package org.jcodec.codecs.mjpeg;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;

/**
 * This class is part of JCodec ( www.jcodec.org )
 * This software is distributed under FreeBSD License
 * 
 * Class for convenient reading JPEG entropy coded bit stream
 * 
 * @author stan
 * 
 */
public class JPEGBitStream {
    InputStream in;
    HuffTable huffYDC;
    HuffTable huffCDC;
    HuffTable huffYAC;
    HuffTable huffCAC;

    /**
     * @deprecated
     * @param coded
     */
    public JPEGBitStream(CodedImage coded) {
        this(coded, new ByteArrayInputStream(coded.getData()));
    }

    public JPEGBitStream(CodedImage coded, InputStream in) {
        this.in = in;
        this.huffYDC = coded.getYdc();
        this.huffCDC = coded.getCdc();
        this.huffYAC = coded.getYac();
        this.huffCAC = coded.getCac();
    }

    private final int prevLumDC[] = new int[1];
    private final int prevCbDC[] = new int[1];
    private final int prevCrDC[] = new int[1];

    public MCU readBlock(MCU block) throws IOException {
        block.clear();
        for (int i = 0; i < block.lum.data.length; i++) {
            readDCValue(block.lum.data[i], prevLumDC, huffYDC);
            readACValues(block.lum.data[i], huffYAC);
        }
        for (int i = 0; i < block.cb.data.length; i++) {
            readDCValue(block.cb.data[i], prevCbDC, huffCDC);
            readACValues(block.cb.data[i], huffCAC);
        }
        for (int i = 0; i < block.cr.data.length; i++) {
            readDCValue(block.cr.data[i], prevCrDC, huffCDC);
            readACValues(block.cr.data[i], huffCAC);
        }
        return block;
    }

    private void bytePad() throws IOException {
        if (bitNo > 0) {
            readValue(8 - bitNo);
        }
    }

    public void readDCValue(int[] target, int[] prevDC, HuffTable table)
            throws IOException {
        int code = readHuff(table);
        if (code != 0) {
            target[0] = readValue(code);
            target[0] += prevDC[0];
            prevDC[0] = target[0];
        } else {
            target[0] = prevDC[0];
        }
    }

    public void readACValues(int[] target, HuffTable table) throws IOException {
        int code;
        int curOff = 1;
        do {
            code = readHuff(table);
            if (code == 0xF0) {
                curOff += 16;
            } else if (code > 0) {
                int rle = code >> 4;
                curOff += rle;
                int len = code & 0xf;
                target[curOff] = readValue(len);
                curOff++;
            }
        } while (code != 0 && curOff < 64);
    }

    private byte currentByte = -1;
    private int bitNo = 8;

    private int readHuff(HuffTable table) throws IOException {
        int c = 0;

        if (bitNo > 7) {
            readNextByte();
        }
        for (int i = 0;; i++) {
            int bit = (((currentByte & 0xff) >> (7 - bitNo)) & 0x1);
            c = (c << 1) + bit;
            bitNo++;
            if (bitNo > 7) {
                if (readNextByte() == -1)
                    return 0;
            }
            if (table.table[c] != -1 && table.lens[c] == (i + 1))
                break;
        }

        return table.table[c];
    }

    private int readNextByte() throws IOException {
        bitNo = 0;
        int b = in.read();
        currentByte = (byte) b;
        return b;
    }

    private int readValue(int length) throws IOException {
        int c = 0;
        for (int i = 0; i < length; i++) {
            c = (c << 1) + (((currentByte) >> (7 - bitNo)) & 0x1);
            bitNo++;
            if (bitNo > 7) {
                if (readNextByte() == -1)
                    throw new IllegalStateException();
            }
        }

        if (length >= 1 && c < (1 << length - 1)) {
            c = -(1 << length) + 1 + c;
        }

        return c;
    }
}