package org.jcodec.codecs.mjpeg;

import static org.jcodec.codecs.mjpeg.JpegUtils.zigzagDecode;

import java.io.IOException;
import java.io.InputStream;
import java.io.PushbackInputStream;

import org.apache.commons.io.IOUtils;
import org.apache.commons.io.input.CountingInputStream;
import org.jcodec.codecs.mjpeg.HuffTable.Type;
import org.jcodec.codecs.mjpeg.tools.Asserts;
import org.jcodec.common.tools.Debug;

/**
 * This class is part of JCodec ( www.jcodec.org )
 * This software is distributed under FreeBSD License
 * 
 * @author stan
 *
 */
public class JpegParser {
    public CodedImage parse(InputStream is) throws IOException {
        CountingInputStream counter = new CountingInputStream(is);
        return parse(new PushbackInputStream(counter, 2), counter);
    }

    public CodedImage parse(PushbackInputStream is, CountingInputStream counter)
            throws IOException {
        CodedImage image = new CodedImage();
        int curQTable = 0;
        while (true) {
            int marker = is.read();
            if (marker == -1)
                return image;
            if (marker == 0)
                continue;
            if (marker != 0xFF)
                throw new RuntimeException("@"
                        + Long.toHexString(counter.getByteCount())
                        + " Marker expected: 0x" + Integer.toHexString(marker));

            int b = is.read();
            Debug.trace("%s", Markers.toString(b));
            switch (b) {
            case Markers.SOF0:
                image.frame = FrameHeader.read(is);
                Debug.trace("    %s", image.frame);
                break;
            case Markers.DHT:
                int len1 = readShort(is);
                CountingInputStream cis = new CountingInputStream(is);
                while (cis.getCount() < len1 - 2) {
                    HuffTable ht = readHuffmanTable(cis);
                    if (ht.type == Type.YDC) {
                        image.setYdc(ht);
                    } else if (ht.type == Type.CDC) {
                        image.setCdc(ht);
                    } else if (ht.type == Type.YAC) {
                        image.setYac(ht);
                    } else {
                        image.setCac(ht);
                    }
                }
                break;
            case Markers.DQT:
                int len4 = readShort(is);
                CountingInputStream cis1 = new CountingInputStream(is);
                while (cis1.getCount() < len4 - 2) {
                    QuantTable quantTable = readQuantTable(cis1);
                    if (curQTable == 0)
                        image.setQuantLum(quantTable);
                    else
                        image.setQuantChrom(quantTable);
                    curQTable++;
                }
                break;
            case Markers.SOS:
                if (image.scan != null) {
                    throw new IllegalStateException(
                            "unhandled - more than one scan header");
                }
                image.scan = ScanHeader.read(is);
                Debug.trace("    %s", image.scan);
                image.setData(readData(is));
                break;
            case Markers.SOI:
                break;
            case Markers.EOI:
                return image;
            case Markers.APP0:
                // int len10 = readShort(is);
                // byte[] id = new byte[4];
                // is.read(id);
                // if (!Arrays.equals(JFIF, id))
                // throw new RuntimeException("Not a JFIF file");
                // is.skip(1);
                //
                // is.skip(2);
                // int units = is.read();
                // int dx = readShort(is);
                // int dy = readShort(is);
                // int tx = is.read();
                // int ty = is.read();
                // is.skip(tx * ty * 3);
                // break;
            case Markers.APP1:
            case Markers.APP2:
            case Markers.APP3:
            case Markers.APP4:
            case Markers.APP5:
            case Markers.APP6:
            case Markers.APP7:
            case Markers.APP8:
            case Markers.APP9:
            case Markers.APPA:
            case Markers.APPB:
            case Markers.APPC:
            case Markers.APPD:
            case Markers.APPE:
            case Markers.APPF:
                int len3 = readShort(is);
                forceSkip(is, len3 - 2);
                break;
            case Markers.DRI:
                /*
                 * Lr: Define restart interval segment length – Specifies the
                 * length of the parameters in the DRI segment shown in Figure
                 * B.9 (see B.1.1.4).
                 */
                int lr = readShort(is);
                // Ri: Restart interval – Specifies the number of MCU in the
                // restart interval.
                int ri = readShort(is);
                Debug.trace("DRI Lr: %d Ri: %d", lr, ri);
                // A DRI marker segment with Ri equal to zero shall disable
                // restart intervals for the following scans.
                Asserts.assertEquals(0, ri);
                break;
            default: {
                throw new IllegalStateException("unhandled marker "
                        + Markers.toString(b));
            }
            }
        }
    }

    private void forceSkip(InputStream is, int len) throws IOException {
        int skipped = 0;
        while (skipped < len) {
            skipped += is.skip(len - skipped);
        }
    }

    /**
     * Decodes huffman tables
     * 
     * @param orig
     * @return
     * @throws IOException
     */
    private HuffTable readHuffmanTable(InputStream is) throws IOException {

        int tableNo = is.read();
        int[] levelSizes = new int[16];
        for (int i = 0; i < 16; i++) {
            levelSizes[i] = is.read();
        }

        return readTable(is, levelSizes, tableNo);
    }

    private HuffTable readTable(InputStream is, int[] levelSizes,
            int tableNumber) throws IOException {

        int nLevels = 0;
        short[][] levels = new short[16][];
        for (int i = 0; i < 16; i++) {
            int length = levelSizes[i];
            if (length > 0) {
                levels[i] = new short[levelSizes[i]];
                for (int j = 0; j < levelSizes[i]; j++) {
                    levels[i][j] = (short) is.read();
                }
            }
        }

        for (int i = 15; i >= 0; i--) {
            if (levelSizes[i] != 0) {
                nLevels = i + 1;
                break;
            }
        }

        short[] codes = new short[1 << nLevels];
        byte[] lens = new byte[1 << nLevels];
        for (int i = 0; i < (1 << nLevels); i++)
            codes[i] = -1;
        int levelStart = 0;
        for (int i = 0; i < levels.length; i++) {
            if (levels[i] == null) {
                levelStart <<= 1;
                continue;
            }
            int j;
            for (j = 0; j < levels[i].length; j++) {
                int code = (levelStart << 1) + j;
                codes[code] = levels[i][j];
                lens[code] = (byte) (i + 1);
            }
            levelStart = (levelStart << 1) + j;
        }

        return new HuffTable(tableNumber, codes, lens);
    }

    private QuantTable readQuantTable(InputStream is) throws IOException {
        int ind = is.read();

        int[] result = new int[64];
        for (int i = 0; i < 64; i++) {
            result[i] = is.read();
        }
        return new QuantTable(ind, zigzagDecode(result));
    }

    private byte[] readData(PushbackInputStream is) throws IOException {
        return IOUtils.toByteArray(new ByteStuffingInputStream(is));
    }

    private int readShort(InputStream is) throws IOException {
        int b1 = is.read();
        int b2 = is.read();

        return (b1 << 8) + b2;
    }

}
