package org.jcodec.codecs.prores;

import static java.lang.Math.min;
import static java.util.Arrays.fill;
import static org.jcodec.codecs.prores.ProresConsts.dcCodebooks;
import static org.jcodec.codecs.prores.ProresConsts.firstDCCodebook;
import static org.jcodec.codecs.prores.ProresConsts.interlaced_scan;
import static org.jcodec.codecs.prores.ProresConsts.levCodebooks;
import static org.jcodec.codecs.prores.ProresConsts.progressive_scan;
import static org.jcodec.codecs.prores.ProresConsts.runCodebooks;
import static org.jcodec.common.dct.SimpleIDCT10Bit.idct10;
import static org.jcodec.common.io.BitstreamReader.wrap;
import static org.jcodec.common.model.Picture.create422;
import static org.jcodec.common.tools.MathUtil.log2;

import java.io.DataInput;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

import org.jcodec.common.io.InBits;
import org.jcodec.common.model.Picture;
import org.jcodec.common.model.Size;

/**
 * 
 * This class is part of JCodec ( www.jcodec.org ) This software is distributed
 * under FreeBSD License
 * 
 * Decoder for Apple ProRes format
 * 
 * As posted at http://git.videolan.org/?p=ffmpeg.git;a=commitdiff;h=5554d
 * e13b29b9bb812ee5cfd606349873ddf0945
 * 
 * @author Stanislav Vitvitskiy
 * 
 */
public class ProresDecoder {

    private List<String> messages = new ArrayList<String>();;

    public ProresDecoder() {
    }

    private int readCodeword(InBits reader, Codebook codebook) throws IOException {
        int q = 0;
        while (reader.read1Bit() == 0)
            q++;

        int val;
        if (q > codebook.switchBits) {
            int bits = codebook.expOrder - codebook.switchBits + q - 1;
            if (bits + q + 1 > 32) {
                throw new RuntimeException("Codeword damaged");
            }
            val = (int) reader.readNBit(bits) | (1 << bits); // Read value
            val -= (1 << codebook.expOrder); // Offset to zero
            // Offset to next number after max rice
            val += ((codebook.switchBits + 1) << codebook.riceOrder);
        } else if (codebook.riceOrder > 0) {
            val = (q << codebook.riceOrder) + (int) reader.readNBit(codebook.riceOrder);
        } else {
            val = q;
        }
        return val;
    }

    private int golumbToSigned(int val) {
        return (val >> 1) ^ golumbSign(val);
    }

    private int golumbSign(int val) {
        return -(val & 1);
    }

    private int toSigned(int val, int sign) {
        return (val ^ sign) - sign;
    }

    void readDCCoeffs(InBits bits, int[] qMat, int[] out, int blocksPerSlice) throws IOException {
        int c = readCodeword(bits, firstDCCodebook);
        if (c < 0) {
            throw new RuntimeException("DC tex damaged");
        }
        int prevDc = golumbToSigned(c);
        out[0] = 4096 + qScale(qMat, 0, prevDc);

        int code = 5, sign = 0, idx = 64;
        for (int i = 1; i < blocksPerSlice; i++, idx += 64) {
            code = readCodeword(bits, dcCodebooks[min(code, 6)]);
            if (code < 0) {
                throw new RuntimeException("DC tex damaged");
            }

            if (code != 0)
                sign ^= golumbSign(code);
            else
                sign = 0;

            prevDc += toSigned((code + 1) >> 1, sign);
            out[idx] = 4096 + qScale(qMat, 0, prevDc);
        }
    }

    void readACCoeffs(InBits bits, int[] qMat, int[] out, int blocksPerSlice, int[] scan) throws IOException {
        int run = 4;
        int level = 2;

        int blockMask = blocksPerSlice - 1; // since blocksPerSlice is 1 << n
        int log2BlocksPerSlice = log2(blocksPerSlice);
        int maxCoeffs = 64 << log2BlocksPerSlice;

        int pos = blockMask;
        while (bits.moreData()) {
            run = readCodeword(bits, runCodebooks[min(run, 15)]);
            if (run < 0 || run >= maxCoeffs - pos - 1) {
                throw new RuntimeException("AC tex damaged, RUN");
            }
            pos += run + 1;

            level = readCodeword(bits, levCodebooks[min(level, 9)]) + 1;
            if (level < 0 || level > 65535) {
                throw new RuntimeException("DC tex damaged, LEV");
            }
            int sign = -bits.read1Bit();
            int ind = pos >> log2BlocksPerSlice;
            out[((pos & blockMask) << 6) + scan[ind]] = qScale(qMat, ind, toSigned(level, sign));
        }
    }

    private int qScale(int[] qMat, int ind, int val) {
        return ((val * qMat[ind]) >> 2);
    }

    int[] decodeOnePlane(InBits bits, int blocksPerSlice, int[] qMat, int[] scan, int mbX, int mbY, int plane)
            throws IOException {
        int[] out = new int[blocksPerSlice << 6];
        try {
            readDCCoeffs(bits, qMat, out, blocksPerSlice);
            readACCoeffs(bits, qMat, out, blocksPerSlice, scan);
        } catch (RuntimeException e) {
            int x = mbX << 4;
            int y = mbY << 4;
            messages.add(e.getMessage() + " at (" + x + ", " + y + ", " + (x + (blocksPerSlice << 2)) + ", " + (y + 16)
                    + "), plane " + plane);
        }

        for (int i = 0; i < blocksPerSlice; i++) {
            idct10(out, i << 6);
        }

        return out;
    }

    public Picture decodeFrame(InputStream is) throws IOException {
        messages.clear();
        
        DataInput inp = new DataInputStream(is);

        int frameSize = inp.readInt();
        String sig = readSig(inp);
        if (!"icpf".equals(sig))
            throw new RuntimeException("Stan");

        short hdrSize = inp.readShort();
        short version = inp.readShort();

        int res1 = inp.readInt();

        short width = inp.readShort();
        short height = inp.readShort();

        int flags1 = inp.readByte();

        int frameType = (flags1 >> 2) & 3;

        int[] scan;
        boolean topFieldFirst = false;

        if (frameType == 0) {
            scan = progressive_scan;
        } else {
            scan = interlaced_scan;
            if (frameType == 1)
                topFieldFirst = true;
        }

        byte res2 = inp.readByte();
        byte prim = inp.readByte();
        byte transFunc = inp.readByte();
        byte matrix = inp.readByte();
        byte pixFmt = inp.readByte();
        byte res3 = inp.readByte();

        int flags2 = inp.readByte() & 0xff;

        int[] qMatLuma = new int[64];
        int[] qMatChroma = new int[64];

        if (hasQMatLuma(flags2)) {
            readQMat(inp, qMatLuma, scan);
        } else {
            fill(qMatLuma, 4);
        }

        if (hasQMatChroma(flags2)) {
            readQMat(inp, qMatChroma, scan);
        } else {
            fill(qMatChroma, 4);
        }

        inp.skipBytes(hdrSize - (20 + (hasQMatLuma(flags2) ? 64 : 0) + (hasQMatChroma(flags2) ? 64 : 0)));

        if (frameType == 0) {
            return decodePicture(inp, width, height, qMatLuma, qMatChroma, scan);
        } else {
            Picture top = decodePicture(inp, width, height >> 1, qMatLuma, qMatChroma, scan);
            int from = append(messages, ", field 1", 0);
            
            Picture bottom = decodePicture(inp, width, height >> 1, qMatLuma, qMatChroma, scan);
            append(messages, ", field 2", from);
            
            if (topFieldFirst)
                return combine(top, bottom);
            else
                return combine(bottom, top);
        }
    }

    private int append(List<String> mesg, String suff, int from) {
        for(int i = from; i < mesg.size(); i++) {
            mesg.set(i, mesg.get(i) + suff);
        }
        return mesg.size();
    }

    public List<String> getMessages() {
        return new ArrayList<String>(messages);
    }

    private Picture combine(Picture top, Picture bottom) {
        int[] y = new int[top.getY().length << 1];
        int[] cb = new int[top.getCb().length << 1];
        int[] cr = new int[top.getCr().length << 1];
        combinePlane(top.getY(), bottom.getY(), y, top.getWidth(), top.getHeight() << 1);
        combinePlane(top.getCb(), bottom.getCb(), cb, top.getWidth(), top.getHeight());
        combinePlane(top.getCr(), bottom.getCr(), cr, top.getWidth(), top.getHeight());

        return new Picture(top.getWidth(), top.getHeight() << 1, y, cb, cr);
    }

    private void combinePlane(int[] top, int[] bottom, int[] y, int width, int height) {
        height >>= 1;
        int off = 0;
        for (int i = 0; i < height; i++) {
            System.arraycopy(top, off, y, off << 1, width);
            System.arraycopy(bottom, off, y, (off << 1) + width, width);
            off += width;
        }
    }

    private String readSig(DataInput inp) throws IOException {
        byte[] sig = new byte[4];
        inp.readFully(sig);
        return new String(sig);
    }

    private Picture decodePicture(DataInput inp, int width, int height, int[] qMatLuma, int[] qMatChroma, int[] scan)
            throws IOException {
        int hdrSize = (inp.readByte() & 0xff) >> 3;
        inp.skipBytes(4);
        int sliceCount = inp.readShort();

        int a = inp.readByte() & 0xff;
        int log2SliceMbWidth = a >> 4;

        int mbWidth = (width + 15) >> 4;
        int mbHeight = (height + 15) >> 4;

        int sliceMbCount = 1 << log2SliceMbWidth;
        int mbX = 0;
        int mbY = 0;
        short[] sliceSizes = new short[sliceCount];
        for (int i = 0; i < sliceCount; i++) {
            sliceSizes[i] = inp.readShort();
        }

        Picture result = create422(new Size(mbWidth << 4, mbHeight << 4));

        for (int i = 0; i < sliceCount; i++) {

            while (mbWidth - mbX < sliceMbCount)
                sliceMbCount >>= 1;

            decodeSlice(inp, qMatLuma, qMatChroma, scan, sliceMbCount, mbX, mbY, sliceSizes[i], result);

            mbX += sliceMbCount;
            if (mbX == mbWidth) {
                sliceMbCount = 1 << log2SliceMbWidth;
                mbX = 0;
                mbY++;
            }
        }

        return result;
    }

    private void decodeSlice(DataInput inp, int[] qMatLuma, int[] qMatChroma, int[] scan, int sliceMbCount, int mbX,
            int mbY, short sliceSize, Picture result) throws IOException {
        int hdrSize = (inp.readByte() & 0xff) >> 3;
        int qScale = clip(inp.readByte() & 0xff, 1, 224);
        qScale = qScale > 128 ? qScale - 96 << 2 : qScale;
        int yDataSize = inp.readShort();
        int uDataSize = inp.readShort();
        int vDataSize = sliceSize - uDataSize - yDataSize - hdrSize;

        int[] y = decodeOnePlane(bitstream(inp, yDataSize), sliceMbCount << 2, scaleMat(qMatLuma, qScale), scan, mbX,
                mbY, 0);
        int[] u = decodeOnePlane(bitstream(inp, uDataSize), sliceMbCount << 1, scaleMat(qMatChroma, qScale), scan, mbX,
                mbY, 1);
        int[] v = decodeOnePlane(bitstream(inp, vDataSize), sliceMbCount << 1, scaleMat(qMatChroma, qScale), scan, mbX,
                mbY, 2);

        putSlice(result, mbX, mbY, y, u, v);
    }

    private int[] scaleMat(int[] qMatLuma, int qScale) {
        int[] res = new int[qMatLuma.length];
        for (int i = 0; i < qMatLuma.length; i++)
            res[i] = qMatLuma[i] * qScale;

        return res;
    }

    private InBits bitstream(DataInput inp, int yDataSize) throws IOException {
        byte[] sliceData = new byte[yDataSize];
        inp.readFully(sliceData);
        return wrap(sliceData);
    }

    private int clip(int val, int min, int max) {
        return val < min ? min : (val > max ? max : val);
    }

    private void putSlice(Picture result, int mbX, int mbY, int[] y, int[] u, int[] v) {
        int mbPerSlice = y.length >> 8;

        putLuma(result.getY(), result.getWidth(), mbX, mbY, y, mbPerSlice);
        putChroma(result.getCb(), result.getWidth() >> 1, mbX, mbY, u, mbPerSlice);
        putChroma(result.getCr(), result.getWidth() >> 1, mbX, mbY, v, mbPerSlice);
    }

    private void putLuma(int[] y, int stride, int mbX, int mbY, int[] luma, int mbPerSlice) {
        int off = (mbX << 4) + (mbY << 4) * stride;
        for (int k = 0; k < mbPerSlice; k++) {
            putBlock(y, off, stride, luma, k << 8);
            putBlock(y, off + 8, stride, luma, (k << 8) + 64);
            putBlock(y, off + 8 * stride, stride, luma, (k << 8) + 128);
            putBlock(y, off + 8 * stride + 8, stride, luma, (k << 8) + 192);
            off += 16;
        }
    }

    private void putChroma(int[] y, int stride, int mbX, int mbY, int[] chroma, int mbPerSlice) {
        int off = (mbX << 3) + (mbY << 4) * stride;
        for (int k = 0; k < mbPerSlice; k++) {
            putBlock(y, off, stride, chroma, k << 7);
            putBlock(y, off + 8 * stride, stride, chroma, (k << 7) + 64);
            off += 8;
        }
    }

    private void putBlock(int[] square, int sqOff, int sqStride, int[] flat, int flOff) {
        for (int i = 0; i < 8; i++) {
            for (int j = 0; j < 8; j++)
                square[j + sqOff] = clip(flat[j + flOff], 4, 1019);
            sqOff += sqStride;
            flOff += 8;
        }
    }

    private boolean hasQMatChroma(int flags2) {
        return (flags2 & 1) != 0;
    }

    private void readQMat(DataInput inp, int[] qMatLuma, int[] scan) throws IOException {
        byte[] b = new byte[64];
        inp.readFully(b);
        for (int i = 0; i < 64; i++) {
            qMatLuma[i] = b[scan[i]] & 0xff;
        }
    }

    private boolean hasQMatLuma(int flags2) {
        return (flags2 & 2) != 0;
    }
}