package org.jcodec.codecs.prores;

import static java.lang.Math.min;
import static org.jcodec.codecs.prores.ProresConsts.QMAT_CHROMA_APCH;
import static org.jcodec.codecs.prores.ProresConsts.QMAT_CHROMA_APCN;
import static org.jcodec.codecs.prores.ProresConsts.QMAT_CHROMA_APCO;
import static org.jcodec.codecs.prores.ProresConsts.QMAT_CHROMA_APCS;
import static org.jcodec.codecs.prores.ProresConsts.QMAT_LUMA_APCH;
import static org.jcodec.codecs.prores.ProresConsts.QMAT_LUMA_APCN;
import static org.jcodec.codecs.prores.ProresConsts.QMAT_LUMA_APCO;
import static org.jcodec.codecs.prores.ProresConsts.QMAT_LUMA_APCS;
import static org.jcodec.codecs.prores.ProresConsts.dcCodebooks;
import static org.jcodec.codecs.prores.ProresConsts.firstDCCodebook;
import static org.jcodec.codecs.prores.ProresConsts.levCodebooks;
import static org.jcodec.codecs.prores.ProresConsts.progressive_scan;
import static org.jcodec.codecs.prores.ProresConsts.runCodebooks;
import static org.jcodec.common.dct.DCTRef.fdct;
import static org.jcodec.common.tools.ImageOP.subImageWithFill;
import static org.jcodec.common.tools.MathUtil.log2;

import java.io.ByteArrayOutputStream;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;

import org.jcodec.common.io.BitstreamWriter;
import org.jcodec.common.io.OutBits;
import org.jcodec.common.model.Picture;
import org.jcodec.common.model.Plane;
import org.jcodec.common.model.Rect;

/**
 * 
 * This class is part of JCodec ( www.jcodec.org )
 * This software is distributed under FreeBSD License
 * 
 * Apple ProRes encoder
 * @author stan
 *
 */
public class ProresEncoder {

    private static final int DEFAULT_SLICE_MB_WIDTH = 8;

    public static enum Profile {
        PROXY(QMAT_LUMA_APCO, QMAT_CHROMA_APCO, "apco", 1000, 4, 8), LT(QMAT_LUMA_APCS, QMAT_CHROMA_APCS, "apcs", 2100,
                1, 9), STANDARD(
                QMAT_LUMA_APCN, QMAT_CHROMA_APCN, "apcn", 3500, 1, 6), HQ(QMAT_LUMA_APCH, QMAT_CHROMA_APCH, "apch",
                5400, 1, 6);

        final int[] qmatLuma;
        final int[] qmatChroma;
        final public String fourcc;
        // Per 1024 pixels
        final int bitrate;
        final int firstQp;
        final int lastQp;

        private Profile(int[] qmatLuma, int[] qmatChroma, String fourcc, int bitrate, int firstQp, int lastQp) {
            this.qmatLuma = qmatLuma;
            this.qmatChroma = qmatChroma;
            this.fourcc = fourcc;
            this.bitrate = bitrate;
            this.firstQp = firstQp;
            this.lastQp = lastQp;
        }
    };

    private Profile profile;
    private int[][] scaledLuma;
    private int[][] scaledChroma;

    public ProresEncoder(Profile profile) {
        this.profile = profile;
        scaledLuma = scaleQMat(profile.qmatLuma, 1, 16);
        scaledChroma = scaleQMat(profile.qmatChroma, 1, 16);
    }

    private int[][] scaleQMat(int[] qmatLuma, int start, int count) {
        int[][] result = new int[count][];
        for (int i = 0; i < count; i++) {
            result[i] = new int[qmatLuma.length];
            for (int j = 0; j < qmatLuma.length; j++)
                result[i][j] = qmatLuma[j] * (i + start);
        }
        return result;
    }

    private void writeCodeword(OutBits writer, Codebook codebook, int val) throws IOException {
        int firstExp = ((codebook.switchBits + 1) << codebook.riceOrder);
        if (val >= firstExp) {
            val -= firstExp;
            val += (1 << codebook.expOrder); // Offset to zero

            int exp = log2(val);
            int zeros = exp - codebook.expOrder + codebook.switchBits + 1;
            for (int i = 0; i < zeros; i++)
                writer.write1Bit(0);
            writer.write1Bit(1);
            writer.writeNBit(val, exp);

        } else if (codebook.riceOrder > 0) {
            for (int i = 0; i < (val >> codebook.riceOrder); i++)
                writer.write1Bit(0);
            writer.write1Bit(1);
            writer.writeNBit(val & ((1 << codebook.riceOrder) - 1), codebook.riceOrder);
        } else {
            for (int i = 0; i < val; i++)
                writer.write1Bit(0);
            writer.write1Bit(1);
        }
    }

    private int qScale(int[] qMat, int ind, int val) {
        return val / qMat[ind];
    }

    private int toGolumb(int val) {
        return (val << 1) ^ (val >> 31);
    }

    private int toGolumb(int val, int sign) {
        if (val == 0)
            return 0;
        return (val << 1) + sign;
    }

    private int diffSign(int val, int sign) {
        return (val >> 31) ^ sign;
    }

    private int isNegative(int val) {
        return ((val >> 31) ^ -1) + 1;
    }

    private int getLevel(int val) {
        int sign = (val >> 31);
        return (val ^ sign) - sign;
    }

    void writeDCCoeffs(OutBits bits, int[] qMat, int[] in, int blocksPerSlice) throws IOException {
        int prevDc = qScale(qMat, 0, in[0] - 16384);
        writeCodeword(bits, firstDCCodebook, toGolumb(prevDc));

        int code = 5, sign = 0, idx = 64;
        for (int i = 1; i < blocksPerSlice; i++, idx += 64) {
            int newDc = qScale(qMat, 0, in[idx] - 16384);
            int delta = newDc - prevDc;
            int newCode = toGolumb(getLevel(delta), diffSign(delta, sign));
            writeCodeword(bits, dcCodebooks[min(code, 6)], newCode);
            code = newCode;
            sign = delta >> 31;
            prevDc = newDc;
        }
    }

    void writeACCoeffs(OutBits bits, int[] qMat, int[] in, int blocksPerSlice, int[] scan) throws IOException {
        int prevRun = 4;
        int prevLevel = 2;

        int run = 0;
        for (int i = 1; i < 64; i++) {
            int indp = scan[i];
            for (int j = 0; j < blocksPerSlice; j++) {
                int val = qScale(qMat, indp, in[(j << 6) + indp]);
                if (val == 0)
                    run++;
                else {
                    writeCodeword(bits, runCodebooks[min(prevRun, 15)], run);
                    prevRun = run;
                    run = 0;
                    int level = getLevel(val);
                    writeCodeword(bits, levCodebooks[min(prevLevel, 9)], level - 1);
                    prevLevel = level;
                    bits.write1Bit(isNegative(val));
                }
            }
        }
    }

    void encodeOnePlane(OutBits bits, int blocksPerSlice, int[] qMat, int[] scan, int[] in) throws IOException {

        writeDCCoeffs(bits, qMat, in, blocksPerSlice);
        writeACCoeffs(bits, qMat, in, blocksPerSlice, scan);
        bits.flush();
    }

    private void dctOnePlane(int blocksPerSlice, int[] in) {
        for (int i = 0; i < blocksPerSlice; i++) {
            fdct(in, i << 6);
        }
    }

    private int encodeSlice(DataOutput out, int[][] scaledLuma, int[][] scaledChroma, int[] scan, int sliceMbCount,
            int mbX, int mbY, Picture result, int prevQp, int mbWidth, int mbHeight, boolean unsafe) throws IOException {

        Picture striped = splitSlice(result, mbX, mbY, sliceMbCount, unsafe);
        dctOnePlane(sliceMbCount << 2, striped.getY());
        dctOnePlane(sliceMbCount << 1, striped.getCb());
        dctOnePlane(sliceMbCount << 1, striped.getCr());

        int est = (sliceMbCount >> 2) * profile.bitrate;
        int low = est - (est >> 3); // 12% bitrate fluctuation
        int high = est + (est >> 3);

        int qp = prevQp;
        byte[][] data = encodeSliceData(scaledLuma[qp - 1], scaledChroma[qp - 1], scan, sliceMbCount, striped, qp);
        if (bits(data) > high && qp < profile.lastQp) {
            do {
                ++qp;
                data = encodeSliceData(scaledLuma[qp - 1], scaledChroma[qp - 1], scan, sliceMbCount, striped, qp);
            } while (bits(data) > high && qp < profile.lastQp);
        } else if (bits(data) < low && qp > profile.firstQp) {
            do {
                --qp;
                data = encodeSliceData(scaledLuma[qp - 1], scaledChroma[qp - 1], scan, sliceMbCount, striped, qp);
            } while (bits(data) < low && qp > profile.firstQp);
        }

        out.write(6 << 3); // hdr size
        out.write(qp); // qscale
        out.writeShort(data[0].length);
        out.writeShort(data[1].length);

        out.write(data[0]);
        out.write(data[1]);
        out.write(data[2]);

        return qp;
    }

    private int bits(byte[][] data) {
        return (data[0].length + data[1].length + data[2].length) << 3;
    }

    private byte[][] encodeSliceData(int[] qmatLuma, int[] qmatChroma, int[] scan,
            int sliceMbCount, Picture striped, int qp) throws IOException {
        ByteArrayOutputStream bytes1 = new ByteArrayOutputStream();
        encodeOnePlane(new BitstreamWriter(bytes1), sliceMbCount << 2, qmatLuma, scan, striped.getY());
        ByteArrayOutputStream bytes2 = new ByteArrayOutputStream();
        encodeOnePlane(new BitstreamWriter(bytes2), sliceMbCount << 1, qmatChroma, scan, striped.getCb());
        ByteArrayOutputStream bytes3 = new ByteArrayOutputStream();
        encodeOnePlane(new BitstreamWriter(bytes3), sliceMbCount << 1, qmatChroma, scan, striped.getCr());

        return new byte[][] { bytes1.toByteArray(), bytes2.toByteArray(), bytes3.toByteArray() };
    }

    private byte[][] encodePicture(int[][] scaledLuma, int[][] scaledChroma, int[] scan, Picture picture)
            throws IOException {

        int mbWidth = (picture.getWidth() + 15) >> 4;
        int mbHeight = (picture.getHeight() + 15) >> 4;
        int qp = profile.firstQp;

        ByteArrayOutputStream buffer = new ByteArrayOutputStream();
        DataOutputStream dout = new DataOutputStream(buffer);
        ArrayList<Integer> sliceSizes = new ArrayList<Integer>();
        for (int mbY = 0; mbY < mbHeight; mbY++) {
            int mbX = 0;
            int sliceMbCount = DEFAULT_SLICE_MB_WIDTH;
            while (mbX < mbWidth) {
                while (mbWidth - mbX < sliceMbCount)
                    sliceMbCount >>= 1;

                int sliceStart = buffer.size();
                boolean unsafeBottom = (picture.getHeight() % 16) != 0 && mbY == mbHeight - 1;
                boolean unsafeRight = (picture.getWidth() % 16) != 0 && mbX + sliceMbCount == mbWidth;
                qp = encodeSlice(dout, scaledLuma, scaledChroma, scan, sliceMbCount, mbX, mbY, picture, qp, mbWidth,
                        mbHeight, unsafeBottom || unsafeRight);
                sliceSizes.add(buffer.size() - sliceStart);

                mbX += sliceMbCount;
            }
        }

        ByteArrayOutputStream header = new ByteArrayOutputStream();
        DataOutputStream out = new DataOutputStream(header);
        int headerLen = 8;
        out.write(headerLen << 3);
        out.writeInt(0);
        out.writeShort(sliceSizes.size());

        out.writeByte(log2(DEFAULT_SLICE_MB_WIDTH) << 4);

        for (Integer size : sliceSizes) {
            out.writeShort(size);
        }

        return new byte[][] { header.toByteArray(), buffer.toByteArray() };
    }

    private Picture splitSlice(Picture result, int mbX, int mbY, int sliceMbCount, boolean unsafe) {
        if (unsafe) {
            int[] filledY = subImageWithFill(new Plane(result.getY(), result.getYSize()), new Rect(mbX << 4,
                    mbY << 4, sliceMbCount << 4, 16));
            int[] filledCb = subImageWithFill(new Plane(result.getCb(), result.getCbSize()), new Rect(mbX << 3,
                    mbY << 4, sliceMbCount << 3, 16));
            int[] filledCr = subImageWithFill(new Plane(result.getCr(), result.getCrSize()), new Rect(mbX << 3,
                    mbY << 4, sliceMbCount << 3, 16));

            return split(new Picture(sliceMbCount << 4, 16, filledY, filledCb, filledCr), 0, 0, sliceMbCount);
        } else {
            return split(result, mbX, mbY, sliceMbCount);
        }
    }

    private Picture split(Picture result, int mbX, int mbY, int sliceMbCount) {
        return new Picture(sliceMbCount << 4, 16,
                split(result.getY(), result.getYSize().getWidth(), mbX, mbY, sliceMbCount, 0),
                split(result.getCb(), result.getCbSize().getWidth(), mbX, mbY, sliceMbCount, 1),
                split(result.getCr(), result.getCrSize().getWidth(), mbX, mbY, sliceMbCount, 1));
    }

    private int[] split(int[] y, int stride, int mbX, int mbY, int sliceMbCount, int chroma) {
        int out[] = new int[sliceMbCount << (8 - chroma)];

        int outOff = 0;
        int off = (mbY << 4) * stride + (mbX << (4 - chroma));

        for (int i = 0; i < sliceMbCount; i++) {
            splitBlock(y, stride, off, out, outOff);
            splitBlock(y, stride, off + (stride << 3), out, outOff + (128 >> chroma));

            if (chroma == 0) {
                splitBlock(y, stride, off + 8, out, outOff + 64);
                splitBlock(y, stride, off + (stride << 3) + 8, out, outOff + 192);
            }

            outOff += (256 >> chroma);
            off += (16 >> chroma);
        }

        return out;
    }

    private void splitBlock(int[] y, int stride, int off, int[] out, int outOff) {
        for (int i = 0; i < 8; i++) {
            for (int j = 0; j < 8; j++)
                out[outOff++] = y[off++];
            off += stride - 8;
        }
    }

    public void encodeFrame(OutputStream out, Picture frame) throws IOException {

        byte[][] picture = encodePicture(scaledLuma, scaledChroma, progressive_scan, frame);

        DataOutput outp = new DataOutputStream(out);

        int headerSize = 148;
        outp.writeInt(headerSize + 8 + picture[0].length + picture[1].length); // frame size
        outp.write(new byte[] { 'i', 'c', 'p', 'f' });

        outp.writeShort(headerSize); // header size
        outp.writeShort(0);

        outp.write(new byte[] { 'a', 'p', 'l', '0' });

        outp.writeShort(frame.getWidth());
        outp.writeShort(frame.getHeight());

        outp.write(0x83); // {10}(422){00}{00}(frame){11}

        outp.write(new byte[] { 0, 2, 2, 6, 32, 0 });

        outp.write(3); // flags2

        writeQMat(out, profile.qmatLuma);

        writeQMat(out, profile.qmatChroma);

        out.write(picture[0]);
        out.write(picture[1]);
    }

    private void writeQMat(OutputStream out, int[] qmat) throws IOException {
        for (int i = 0; i < 64; i++)
            out.write(qmat[i]);
    }
}