package org.jcodec.containers.mp4;

import static org.jcodec.containers.mp4.boxes.Box.findFirst;

import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.jcodec.common.model.Packet;
import org.jcodec.containers.mp4.boxes.AudioSampleEntry;
import org.jcodec.containers.mp4.boxes.Box;
import org.jcodec.containers.mp4.boxes.ChunkOffsets64Box;
import org.jcodec.containers.mp4.boxes.ChunkOffsetsBox;
import org.jcodec.containers.mp4.boxes.HandlerBox;
import org.jcodec.containers.mp4.boxes.MovieBox;
import org.jcodec.containers.mp4.boxes.NodeBox;
import org.jcodec.containers.mp4.boxes.SampleEntry;
import org.jcodec.containers.mp4.boxes.SampleSizesBox;
import org.jcodec.containers.mp4.boxes.SampleToChunkBox;
import org.jcodec.containers.mp4.boxes.SampleToChunkBox.SampleToChunkEntry;
import org.jcodec.containers.mp4.boxes.SyncSamplesBox;
import org.jcodec.containers.mp4.boxes.TimeToSampleBox;
import org.jcodec.containers.mp4.boxes.TimeToSampleBox.TimeToSampleEntry;
import org.jcodec.containers.mp4.boxes.TimecodeSampleEntry;
import org.jcodec.containers.mp4.boxes.TrakBox;
import org.jcodec.containers.mp4.io.RandomAccessInput;

/**
 * This class is part of JCodec ( www.jcodec.org ) This software is distributed
 * under FreeBSD License
 * 
 * Demuxer frontend for MP4
 * 
 * 
 * @author Stanislav Vitvitskiy
 * 
 */
public class MP4Demuxer {

    private List<DemuxerTrack> tracks;
    private RandomAccessInput input;
    private MovieBox movie;

    public abstract class DemuxerTrack {
        protected TrakBox box;
        private int type;
        private int no;
        protected SampleEntry[] sampleEntries;

        protected TimeToSampleEntry[] timeToSamples;
        protected SampleToChunkEntry[] sampleToChunks;
        protected long[] chunkOffsets;

        protected long duration;

        protected int sttsInd;
        protected int sttsSubInd;

        protected int stcoInd;

        protected int stscInd;

        protected long pts;
        protected int curFrame;

        public DemuxerTrack(TrakBox trak) {
            no = trak.getTrackHeader().getNo();
            type = getTrackType(trak);
            sampleEntries = Box.findAll(trak, SampleEntry.class, "mdia", "minf", "stbl", "stsd", null);

            NodeBox stbl = trak.getMdia().getMinf().getStbl();

            TimeToSampleBox stts = findFirst(stbl, TimeToSampleBox.class, "stts");
            SampleToChunkBox stsc = findFirst(stbl, SampleToChunkBox.class, "stsc");
            ChunkOffsetsBox stco = findFirst(stbl, ChunkOffsetsBox.class, "stco");
            ChunkOffsets64Box co64 = findFirst(stbl, ChunkOffsets64Box.class, "co64");

            timeToSamples = stts.getTimeToSamples();
            sampleToChunks = stsc.getSampleToChunk();
            chunkOffsets = stco != null ? stco.getChunkOffsets() : co64.getChunkOffsets();

            for (TimeToSampleEntry ttse : timeToSamples) {
                duration += ttse.getSampleCount() * ttse.getSampleDuration();
            }

            box = trak;
        }

        public int getType() {
            return type;
        }

        public int getNo() {
            return no;
        }

        public SampleEntry[] getSampleEntries() {
            return sampleEntries;
        }

        public TrakBox getBox() {
            return box;
        }

        public long getTimescale() {
            return box.getTimescale();
        }

        public abstract Packet getFrames(int n) throws IOException;

        protected abstract void seekPointer(int frameNo);

        public boolean seek(long pts) {
            if (pts < 0)
                throw new IllegalArgumentException("Seeking to negative pts");
            if (pts >= duration)
                return false;

            sttsInd = 0;
            long prevDur = 0;
            int frameNo = 0;
            while (pts > prevDur + timeToSamples[sttsInd].getSampleCount() * timeToSamples[sttsInd].getSampleDuration()) {
                prevDur += timeToSamples[sttsInd].getSampleCount() * timeToSamples[sttsInd].getSampleDuration();
                frameNo += timeToSamples[sttsInd].getSampleCount();
            }
            sttsSubInd = (int) ((pts - prevDur) / timeToSamples[sttsInd].getSampleDuration());
            frameNo += sttsSubInd;

            seekPointer(frameNo);

            return true;
        }

        protected long shiftPts(int frames) {
            long result = 0;
            int rem;
            while (frames > (rem = timeToSamples[sttsInd].getSampleCount() - sttsSubInd)) {
                frames -= rem;
                result += rem * timeToSamples[sttsInd].getSampleDuration();
                sttsInd++;
                sttsSubInd = 0;
                if (sttsInd >= timeToSamples.length)
                    return result;
            }
            result += frames * timeToSamples[sttsInd].getSampleDuration();

            pts += result;

            return result;
        }

        protected void nextChunk() {
            if (stcoInd >= chunkOffsets.length)
                return;
            stcoInd++;

            if ((stscInd + 1 < sampleToChunks.length) && stcoInd + 1 == sampleToChunks[stscInd + 1].getFirst()) {
                stscInd++;
            }
        }

        public boolean gotoFrame(int frameNo) {
            if (frameNo < 0)
                throw new IllegalArgumentException("negative frame number");
            if (frameNo >= getFrameCount())
                return false;

            seekPointer(frameNo);
            seekPts(frameNo);

            return true;
        }

        private void seekPts(int frameNo) {
            pts = sttsInd = sttsSubInd = 0;
            shiftPts(frameNo);
        }

        public int getStartTimecode() throws IOException {
            if (!box.isTimecode())
                throw new IllegalStateException("Not a timecode track");
            Packet nextFrame = getFrames(1);
            gotoFrame(0);
            byte[] raw = nextFrame.getData();
            return ((raw[0] & 0xff) << 24) | ((raw[1] & 0xff) << 16) | ((raw[2] & 0xff) << 8) | (raw[3] & 0xff);
        }

        public int parseTimecode(String tc) {
            String[] split = tc.split(":");

            TimecodeSampleEntry tmcd = Box.findFirst(box, TimecodeSampleEntry.class, "mdia", "minf", "stbl", "stsd",
                    "tmcd");
            byte nf = tmcd.getNumFrames();

            return Integer.parseInt(split[3]) + Integer.parseInt(split[2]) * nf + Integer.parseInt(split[1]) * 60 * nf
                    + Integer.parseInt(split[0]) * 3600 * nf;
        }

        public long getDuration() {
            return box.getDuration();
        }

        public abstract long getFrameCount();

        public String formatTimecode(int i) {
            if (!box.isTimecode())
                throw new IllegalStateException("Not a timecode track");
            TimecodeSampleEntry tmcd = Box.findFirst(box, TimecodeSampleEntry.class, "mdia", "minf", "stbl", "stsd",
                    "tmcd");
            byte nf = tmcd.getNumFrames();

            String tc = String.format("%02d", i % nf);
            i /= nf;
            tc = String.format("%02d", i % 60) + ":" + tc;
            i /= 60;
            tc = String.format("%02d", i % 60) + ":" + tc;
            i /= 60;
            tc = String.format("%02d", i) + ":" + tc;

            return tc;
        }
    }

    public DemuxerTrack create(TrakBox trak) {
        SampleSizesBox stsz = findFirst(trak, SampleSizesBox.class, "mdia", "minf", "stbl", "stsz");
        if (stsz.getDefaultSize() == 0)
            return new FramesTrack(trak);
        else
            return new SamplesTrack(trak);
    }

    /**
     * Track of audio samples
     * 
     * @author stan
     * 
     */
    private class SamplesTrack extends DemuxerTrack {

        private int defaultSampleSize;

        private int posShift;

        protected int totalFrames;

        public SamplesTrack(TrakBox trak) {
            super(trak);

            SampleSizesBox stsz = findFirst(trak, SampleSizesBox.class, "mdia", "minf", "stbl", "stsz");
            defaultSampleSize = stsz.getDefaultSize();

            int chunks = 0;
            for (int i = 1; i < sampleToChunks.length; i++) {
                int ch = (int) (sampleToChunks[i].getFirst() - sampleToChunks[i - 1].getFirst());
                totalFrames += ch * sampleToChunks[i - 1].getCount();
                chunks += ch;
            }
            totalFrames += sampleToChunks[sampleToChunks.length - 1].getCount() * (chunkOffsets.length - chunks);
        }

        public Packet getFrames(int n) throws IOException {
            if (n < 0)
                throw new IllegalArgumentException("Negative number of frames");

            int frameSize = getFrameSize();
            n = curFrame + n > totalFrames ? totalFrames - curFrame : n;
            if (n == 0) {
                return null;
            }

            byte[] result = new byte[frameSize * n];

            int rOff = 0;
            do {
                int chSize = sampleToChunks[stscInd].getCount() * frameSize;
                int toRead = result.length - rOff > chSize - posShift ? chSize - posShift : result.length - rOff;
                input.seek(chunkOffsets[stcoInd] + posShift);
                int read = input.read(result, rOff, toRead);
                rOff += read;
                posShift += read;

                if (posShift == chSize) {
                    nextChunk();
                    posShift = 0;
                }
            } while (rOff < result.length);

            long _pts = pts;
            shiftPts(n);

            Packet pkt = new Packet(result, _pts, (int) (pts - _pts), curFrame, true);

            curFrame += n;

            return pkt;
        }

        private int getFrameSize() {
            SampleEntry entry = sampleEntries[sampleToChunks[stscInd].getEntry() - 1];
            if (entry instanceof AudioSampleEntry) {
                return ((AudioSampleEntry) entry).calcSampleSize();
            } else {
                return defaultSampleSize;
            }
        }

        protected void seekPointer(int frameNo) {
            for (stcoInd = 0, stscInd = 0, curFrame = 0;;) {
                int nextFrame = curFrame + sampleToChunks[stscInd].getCount();
                if (nextFrame > frameNo)
                    break;
                curFrame = nextFrame;
                nextChunk();
            }
            posShift = (frameNo - curFrame) * getFrameSize();
        }

        public long getFrameCount() {
            return totalFrames;
        }
    }

    /**
     * Track of video frames
     * 
     * @author stan
     * 
     */
    public class FramesTrack extends DemuxerTrack {

        private int[] sizes;

        private int offInChunk;

        private int noInChunk;

        private int[] syncSamples;
        private int ssOff;

        public FramesTrack(TrakBox trak) {
            super(trak);

            SampleSizesBox stsz = findFirst(trak, SampleSizesBox.class, "mdia", "minf", "stbl", "stsz");
            SyncSamplesBox stss = Box.findFirst(trak, SyncSamplesBox.class, "mdia", "minf", "stbl", "stss");
            if (stss != null) {
                syncSamples = stss.getSyncSamples();
            }
            sizes = stsz.getSizes();
        }

        public Packet getFrames(int n) throws IOException {
            if (n != 1)
                throw new IllegalArgumentException("Frames should be = 1 for this track");

            if (curFrame >= sizes.length)
                return null;
            int size = sizes[curFrame];

            byte[] result = new byte[size];
            input.seek(chunkOffsets[stcoInd] + offInChunk);
            input.read(result);

            int duration = timeToSamples[sttsInd].getSampleDuration();

            boolean sync = syncSamples == null;

            if (syncSamples != null && ssOff < syncSamples.length && curFrame == syncSamples[ssOff]) {
                sync = true;
                ssOff++;
            }

            Packet pkt = new Packet(result, pts, duration, curFrame, sync);

            offInChunk += size;

            curFrame++;
            noInChunk++;
            sttsSubInd++;
            if (noInChunk >= sampleToChunks[stscInd].getCount()) {
                noInChunk = 0;
                offInChunk = 0;

                nextChunk();
            }
            shiftPts(1);

            return pkt;
        }

        protected void seekPointer(int frameNo) {
            curFrame = frameNo;
            stcoInd = 0;
            stscInd = 0;
            noInChunk = frameNo;
            offInChunk = 0;

            while (noInChunk >= sampleToChunks[stscInd].getCount()) {
                noInChunk -= sampleToChunks[stscInd].getCount();

                nextChunk();
            }

            for (int i = 0; i < noInChunk; i++) {
                offInChunk += sizes[frameNo - noInChunk + i];
            }
        }

        public long getFrameCount() {
            return sizes.length;
        }
    };

    public MP4Demuxer(RandomAccessInput input) throws IOException, MP4DemuxerException {
        this.input = input;
        tracks = new LinkedList<DemuxerTrack>();
        findMovieBox(input);
    }

    public DemuxerTrack[] getTracks() {
        return tracks.toArray(new DemuxerTrack[] {});
    }

    private void findMovieBox(RandomAccessInput input) throws IOException, MP4DemuxerException {
        movie = MP4Util.parseMovie(input);
        if (movie == null)
            throw new MP4DemuxerException("Could not find movie meta information box");

        processHeader(movie);
    }

    private void processHeader(NodeBox moov) throws IOException, MP4DemuxerException {
        for (TrakBox trak : Box.findAll(moov, TrakBox.class, "trak")) {
            tracks.add(create(trak));
        }
    }

    public static int getTrackType(TrakBox trak) {
        HandlerBox handler = findFirst(trak, HandlerBox.class, "mdia", "hdlr");
        if (handler == null)
            return -1;
        if ("vide".equals(handler.getComponentSubType()))
            return 0;
        else if ("soun".equals(handler.getComponentSubType()))
            return 1;
        else if ("hint".equals(handler.getComponentSubType()))
            return 2;
        return -1;
    }

    public DemuxerTrack getVideoTrack() {
        for (DemuxerTrack demuxerTrack : tracks) {
            if (demuxerTrack.box.isVideo())
                return demuxerTrack;
        }
        return null;
    }

    public MovieBox getMovie() {
        return movie;
    }

    public DemuxerTrack getTrack(int no) {
        return tracks.get(no);
    }

    public List<DemuxerTrack> getAudioTracks() {
        ArrayList<DemuxerTrack> result = new ArrayList<DemuxerTrack>();
        for (DemuxerTrack demuxerTrack : tracks) {
            if (demuxerTrack.box.isAudio())
                result.add(demuxerTrack);
        }
        return result;
    }

    public DemuxerTrack getTimecodeTrack() {
        for (DemuxerTrack demuxerTrack : tracks) {
            if (demuxerTrack.box.isTimecode())
                return (DemuxerTrack) demuxerTrack;
        }
        return null;
    }
}