package org.jcodec.containers.mp4;

import static org.jcodec.containers.mp4.TrackType.SOUND;
import static org.jcodec.containers.mp4.TrackType.TIMECODE;
import static org.jcodec.containers.mp4.TrackType.VIDEO;
import gnu.trove.list.array.TIntArrayList;
import gnu.trove.list.array.TLongArrayList;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

import javax.sound.sampled.AudioFormat;

import junit.framework.Assert;

import org.jcodec.common.model.Rational;
import org.jcodec.common.model.RationalLarge;
import org.jcodec.common.model.Size;
import org.jcodec.common.model.Unit;
import org.jcodec.containers.mp4.boxes.AudioSampleEntry;
import org.jcodec.containers.mp4.boxes.Box;
import org.jcodec.containers.mp4.boxes.ChunkOffsetsBox;
import org.jcodec.containers.mp4.boxes.DataInfoBox;
import org.jcodec.containers.mp4.boxes.DataRefBox;
import org.jcodec.containers.mp4.boxes.EndianBox;
import org.jcodec.containers.mp4.boxes.EndianBox.Endian;
import org.jcodec.containers.mp4.boxes.FileTypeBox;
import org.jcodec.containers.mp4.boxes.FormatBox;
import org.jcodec.containers.mp4.boxes.GenericMediaInfoBox;
import org.jcodec.containers.mp4.boxes.HandlerBox;
import org.jcodec.containers.mp4.boxes.Header;
import org.jcodec.containers.mp4.boxes.LeafBox;
import org.jcodec.containers.mp4.boxes.MediaBox;
import org.jcodec.containers.mp4.boxes.MediaHeaderBox;
import org.jcodec.containers.mp4.boxes.MediaInfoBox;
import org.jcodec.containers.mp4.boxes.MovieBox;
import org.jcodec.containers.mp4.boxes.MovieHeaderBox;
import org.jcodec.containers.mp4.boxes.NodeBox;
import org.jcodec.containers.mp4.boxes.PixelAspectExt;
import org.jcodec.containers.mp4.boxes.SampleDescriptionBox;
import org.jcodec.containers.mp4.boxes.SampleEntry;
import org.jcodec.containers.mp4.boxes.SampleSizesBox;
import org.jcodec.containers.mp4.boxes.SampleToChunkBox;
import org.jcodec.containers.mp4.boxes.SampleToChunkBox.SampleToChunkEntry;
import org.jcodec.containers.mp4.boxes.SoundMediaHeaderBox;
import org.jcodec.containers.mp4.boxes.SyncSamplesBox;
import org.jcodec.containers.mp4.boxes.TimeToSampleBox;
import org.jcodec.containers.mp4.boxes.TimeToSampleBox.TimeToSampleEntry;
import org.jcodec.containers.mp4.boxes.TimecodeMediaInfoBox;
import org.jcodec.containers.mp4.boxes.TrackHeaderBox;
import org.jcodec.containers.mp4.boxes.TrakBox;
import org.jcodec.containers.mp4.boxes.VideoMediaHeaderBox;
import org.jcodec.containers.mp4.boxes.VideoSampleEntry;

/**
 * This class is part of JCodec ( www.jcodec.org ) This software is distributed
 * under FreeBSD License
 * 
 * Creates MP4 file out of a set of samples
 * 
 * @author Stanislav Vitvitskiy
 * 
 */
public class MP4Muxer {
    private List<MuxerTrack> tracks = new ArrayList<MuxerTrack>();
    private long mdatOffset;
    private long fileOffset;

    private int nextTrackId = 1;
    private RandomAccessFile out;

    public MP4Muxer(RandomAccessFile output) throws IOException {
        this.out = output;
        FileTypeBox ftyp = new FileTypeBox("qt  ", 0x20050300, Arrays.asList(new String[] { "qt  " }));
        ftyp.write(out);
        new Header("wide", 8).write(output);
        mdatOffset = ftyp.getHeader().getBodySize() + 16;
        new Header("mdat", 0).write(output);
        fileOffset = mdatOffset + 8;
    }

    public CompressedTrack addVideoTrack(String fourcc, Size size, String encoderName, int timescale) {
        CompressedTrack track = addTrackForCompressed(VIDEO, timescale);

        track.addSampleEntry(videoSampleEntry(fourcc, size, encoderName));
        return track;
    }

    public VideoSampleEntry videoSampleEntry(String fourcc, Size size, String encoderName) {
        return new VideoSampleEntry(new Header(fourcc), (short) 0, (short) 0, "jcod", 0, 768, size.getWidth(),
                size.getHeight(), 72, 72, (short) 1, encoderName != null ? encoderName : "jcodec", (short) 24,
                (short) 1, (short) -1);
    }

    public static AudioSampleEntry audioSampleEntry(String fourcc, int drefId, int sampleSize, int channels,
            int sampleRate, Endian endian) {
        AudioSampleEntry ase = new AudioSampleEntry(new Header(fourcc, 0), (short) drefId, (short) channels,
                (short) 16, sampleRate, (short) 0, 0, 65535, 0, 1, sampleSize, channels * sampleSize, sampleSize,
                (short) 1);

        NodeBox wave = new NodeBox(new Header("wave"));
        ase.add(wave);

        wave.add(new FormatBox(fourcc));
        wave.add(new EndianBox(endian));
        wave.add(terminatorAtom());
        // ase.add(new ChannelBox(atom));

        return ase;
    }

    private static LeafBox terminatorAtom() {
        return new LeafBox(new Header(new String(new byte[4])), new byte[0]);
    }

    public CompressedTrack addTrackForCompressed(TrackType type, int timescale) {
        CompressedTrack track = new CompressedTrack(nextTrackId++, type, timescale);
        tracks.add(track);
        return track;
    }

    public UncompressedTrack addTrackForUncompressed(TrackType type, int timescale, int sampleDuration, int sampleSize,
            SampleEntry se) {
        UncompressedTrack track = new UncompressedTrack(nextTrackId++, type, timescale, sampleDuration, sampleSize, se);
        tracks.add(track);
        return track;
    }

    public List<MuxerTrack> getTracks() {
        return tracks;
    }

    public abstract class MuxerTrack {
        protected int trackId;
        protected TrackType type;
        protected int timescale;

        protected Rational tgtChunkDuration;
        protected Unit tgtChunkDurationUnit;

        protected long chunkDuration;
        protected List<byte[]> curChunk = new ArrayList<byte[]>();

        protected List<SampleToChunkEntry> samplesInChunks = new ArrayList<SampleToChunkEntry>();
        protected int samplesInLastChunk = -1;
        protected int chunkNo = 0;

        protected boolean finished;

        protected List<SampleEntry> sampleEntries = new ArrayList<SampleEntry>();

        public MuxerTrack(int trackId, TrackType type, int timescale) {
            this.trackId = trackId;
            this.type = type;
            this.timescale = timescale;
        }

        public void setTgtChunkDuration(Rational duration, Unit unit) {
            this.tgtChunkDuration = duration;
            this.tgtChunkDurationUnit = unit;
        }

        public abstract long getTrackTotalDuration();

        public int getTimescale() {
            return timescale;
        }

        protected abstract Box finish() throws IOException;

        public boolean isVideo() {
            return type == VIDEO;
        }

        public boolean isTimecode() {
            return type == TIMECODE;
        }

        public boolean isAudio() {
            return type == SOUND;
        }

        public Size getDisplayDimensions() {
            int width = 0, height = 0;
            if (sampleEntries.get(0) instanceof VideoSampleEntry) {
                VideoSampleEntry vse = (VideoSampleEntry) sampleEntries.get(0);
                Rational pasp = Box.findFirst(vse, PixelAspectExt.class, PixelAspectExt.fourcc()).getRational();
                width = (int) (pasp.getNum() * vse.getWidth()) / pasp.getDen();
                height = (int) vse.getHeight();
            }
            return new Size(width, height);
        }

        public void addSampleEntry(SampleEntry se) {
            if (finished)
                throw new IllegalStateException("The muxer track has finished muxing");
            sampleEntries.add(se);
        }

        public List<SampleEntry> getEntries() {
            return sampleEntries;
        }
    }

    public class UncompressedTrack extends MuxerTrack {

        private int frameDuration;
        private int frameSize;
        private int framesInCurChunk;

        private TLongArrayList chunkOffsets = new TLongArrayList();
        private int totalFrames;

        public UncompressedTrack(int trackId, TrackType type, int timescale, int frameDuration, int frameSize,
                SampleEntry se) {
            super(trackId, type, timescale);
            this.frameDuration = frameDuration;
            this.frameSize = frameSize;
            addSampleEntry(se);

            setTgtChunkDuration(new Rational(1, 2), Unit.SEC);
        }

        public void addSamples(byte[] buffer) throws IOException {
            curChunk.add(buffer);

            int frames = buffer.length / frameSize;
            totalFrames += frames;

            framesInCurChunk += frames;
            chunkDuration += frames * frameDuration;

            outChunkIfNeeded();
        }
        
        private void outChunkIfNeeded() throws IOException {
            Assert.assertTrue(tgtChunkDurationUnit == Unit.FRAME || tgtChunkDurationUnit == Unit.SEC);

            if (tgtChunkDurationUnit == Unit.FRAME
                    && framesInCurChunk * tgtChunkDuration.getDen() == tgtChunkDuration.getNum()) {
                outChunk();
            } else if (tgtChunkDurationUnit == Unit.SEC && chunkDuration > 0
                    && chunkDuration * tgtChunkDuration.getDen() >= tgtChunkDuration.getNum() * timescale) {
                outChunk();
            }
        }

        private void outChunk() throws IOException {
            if (framesInCurChunk == 0)
                return;

            chunkOffsets.add(fileOffset);

            int size = 0;
            for (byte[] b : curChunk) {
                out.write(b);
                size += b.length;
            }
            curChunk.clear();
            fileOffset += size;

            if (samplesInLastChunk == -1 || framesInCurChunk != samplesInLastChunk) {
                samplesInChunks.add(new SampleToChunkEntry(chunkNo + 1, framesInCurChunk, 1));
            }
            samplesInLastChunk = framesInCurChunk;

            chunkNo++;

            framesInCurChunk = 0;
            chunkDuration = 0;
        }

        protected Box finish() throws IOException {
            if (finished)
                throw new IllegalStateException("The muxer track has finished muxing");

            outChunk();

            finished = true;

            TrakBox trak = new TrakBox();
            Size dd = getDisplayDimensions();
            TrackHeaderBox tkhd = new TrackHeaderBox(trackId, totalFrames * frameDuration, dd.getWidth(),
                    dd.getHeight(), new Date().getTime(), new Date().getTime(), 1.0f, (short) 0, 0, new int[] {
                            0x10000, 0, 0, 0, 0x10000, 0, 0, 0, 0x40000000 });
            tkhd.setFlags(0xf);
            trak.add(tkhd);

            MediaBox media = new MediaBox();
            trak.add(media);
            media.add(new MediaHeaderBox(timescale, totalFrames * frameDuration, 0, new Date().getTime(), new Date()
                    .getTime(), 0));

            HandlerBox hdlr = new HandlerBox("mhlr", type.getHandler(), "appl", 0, 0);
            media.add(hdlr);

            MediaInfoBox minf = new MediaInfoBox();
            media.add(minf);
            mediaHeader(minf, type);
            minf.add(new HandlerBox("dhlr", "url ", "appl", 0, 0));
            addDref(minf);
            NodeBox stbl = new NodeBox(new Header("stbl"));
            minf.add(stbl);

            stbl.add(new SampleDescriptionBox(sampleEntries.toArray(new SampleEntry[0])));
            stbl.add(new SampleToChunkBox(samplesInChunks.toArray(new SampleToChunkEntry[0])));
            stbl.add(new SampleSizesBox(frameSize, totalFrames));
            stbl.add(new TimeToSampleBox(
                    new TimeToSampleEntry[] { new TimeToSampleEntry(totalFrames, frameDuration) }));
            stbl.add(new ChunkOffsetsBox(chunkOffsets.toArray()));

            return trak;
        }

        public long getTrackTotalDuration() {
            return totalFrames * frameDuration;
        }
    }

    public class CompressedTrack extends MuxerTrack {

        private List<TimeToSampleEntry> sampleDurations = new ArrayList<TimeToSampleEntry>();
        private long sameDurCount = 0;
        private long curDuration = -1;

        private TLongArrayList chunkOffsets = new TLongArrayList();
        private TIntArrayList sampleSizes = new TIntArrayList();
        private TIntArrayList iframes = new TIntArrayList();

        private int lastEntry = -1;

        private long trackTotalDuration;
        private int curFrame;
        private boolean allIframes = true;

        public CompressedTrack(int trackId, TrackType type, int timescale) {
            super(trackId, type, timescale);

            setTgtChunkDuration(new Rational(1, 1), Unit.FRAME);
        }

        public void addFrame(byte[] buf, int entryNo, long duration, boolean iframe) throws IOException {
            if (finished)
                throw new IllegalStateException("The muxer track has finished muxing");

            if (lastEntry != -1 && lastEntry != entryNo) {
                outChunk(lastEntry);
                samplesInLastChunk = -1;
            }

            curChunk.add(buf);

            if (iframe)
                iframes.add(curFrame);
            else
                allIframes = false;

            curFrame++;

            chunkDuration += duration;
            if (curDuration != -1 && duration != curDuration) {
                sampleDurations.add(new TimeToSampleEntry((int) sameDurCount, (int) curDuration));
                sameDurCount = 0;
            }
            curDuration = duration;
            sameDurCount++;
            trackTotalDuration += duration;

            outChunkIfNeeded(entryNo);

            lastEntry = entryNo;
        }

        private void outChunkIfNeeded(int entryNo) throws IOException {
            Assert.assertTrue(tgtChunkDurationUnit == Unit.FRAME || tgtChunkDurationUnit == Unit.SEC);

            if (tgtChunkDurationUnit == Unit.FRAME
                    && curChunk.size() * tgtChunkDuration.getDen() == tgtChunkDuration.getNum()) {
                outChunk(entryNo);
            } else if (tgtChunkDurationUnit == Unit.SEC && chunkDuration > 0
                    && chunkDuration * tgtChunkDuration.getDen() >= tgtChunkDuration.getNum() * timescale) {
                outChunk(entryNo);
            }
        }

        void outChunk(int entryNo) throws IOException {
            if (curChunk.size() == 0)
                return;

            chunkOffsets.add(fileOffset);

            int total = 0;
            for (byte[] bs : curChunk) {
                out.write(bs);
                total += bs.length;
                sampleSizes.add(bs.length);
            }
            fileOffset += total;

            Assert.assertEquals(fileOffset, out.getFilePointer());

            if (samplesInLastChunk == -1 || samplesInLastChunk != curChunk.size()) {
                samplesInChunks.add(new SampleToChunkEntry(chunkNo + 1, curChunk.size(), entryNo));
            }
            samplesInLastChunk = curChunk.size();
            chunkNo++;

            chunkDuration = 0;
            curChunk.clear();
        }

        protected Box finish() throws IOException {
            if (finished)
                throw new IllegalStateException("The muxer track has finished muxing");

            outChunk(lastEntry);

            if (sameDurCount > 0) {
                sampleDurations.add(new TimeToSampleEntry((int) sameDurCount, (int) curDuration));
            }
            finished = true;

            TrakBox trak = new TrakBox();
            Size dd = getDisplayDimensions();
            TrackHeaderBox tkhd = new TrackHeaderBox(trackId, trackTotalDuration, dd.getWidth(), dd.getHeight(),
                    new Date().getTime(), new Date().getTime(), 1.0f, (short) 0, 0, new int[] { 0x10000, 0, 0, 0,
                            0x10000, 0, 0, 0, 0x40000000 });
            tkhd.setFlags(0xf);
            trak.add(tkhd);

            MediaBox media = new MediaBox();
            trak.add(media);
            media.add(new MediaHeaderBox(timescale, trackTotalDuration, 0, new Date().getTime(), new Date().getTime(),
                    0));

            HandlerBox hdlr = new HandlerBox("mhlr", type.getHandler(), "appl", 0, 0);
            media.add(hdlr);

            MediaInfoBox minf = new MediaInfoBox();
            media.add(minf);
            mediaHeader(minf, type);
            minf.add(new HandlerBox("dhlr", "url ", "appl", 0, 0));
            addDref(minf);
            NodeBox stbl = new NodeBox(new Header("stbl"));
            minf.add(stbl);

            stbl.add(new SampleDescriptionBox(sampleEntries.toArray(new SampleEntry[0])));
            stbl.add(new SampleToChunkBox(samplesInChunks.toArray(new SampleToChunkEntry[0])));
            stbl.add(new SampleSizesBox(sampleSizes.toArray()));
            stbl.add(new TimeToSampleBox(sampleDurations.toArray(new TimeToSampleEntry[] {})));
            stbl.add(new ChunkOffsetsBox(chunkOffsets.toArray()));
            if (!allIframes && iframes.size() > 0)
                stbl.add(new SyncSamplesBox(iframes.toArray()));

            return trak;
        }

        public long getTrackTotalDuration() {
            return trackTotalDuration;
        }

        public void addTimecode(SampleEntry sampleEntry, int counter, RationalLarge dur) throws IOException {
            addSampleEntry(sampleEntry);

            byte[] b = new byte[] { (byte) ((counter >> 24) & 0xff), (byte) ((counter >> 16) & 0xff),
                    (byte) ((counter >> 8) & 0xff), (byte) (counter & 0xff) };

            addFrame(b, getEntries().size(), (dur.getNum() * getTimescale()) / dur.getDen(), false);
        }
    }

    public void writeHeader() throws IOException {
        NodeBox movie = new MovieBox();

        for (MuxerTrack track : tracks) {
            movie.add(track.finish());
        }

        long mdatSize = fileOffset - mdatOffset;
        movieHeader(movie);
        movie.write(out);

        out.seek(mdatOffset);
        out.writeInt((int) mdatSize);
    }

    private void mediaHeader(MediaInfoBox minf, TrackType type) {
        switch (type) {
        case VIDEO:
            VideoMediaHeaderBox vmhd = new VideoMediaHeaderBox(0, 0, 0, 0);
            vmhd.setFlags(1);
            minf.add(vmhd);
            break;
        case SOUND:
            SoundMediaHeaderBox smhd = new SoundMediaHeaderBox();
            smhd.setFlags(1);
            minf.add(smhd);
            break;
        case TIMECODE:
            NodeBox gmhd = new NodeBox(new Header("gmhd"));
            gmhd.add(new GenericMediaInfoBox());
            NodeBox tmcd = new NodeBox(new Header("tmcd"));
            gmhd.add(tmcd);
            tmcd.add(new TimecodeMediaInfoBox((short) 0, (short) 0, (short) 12, new short[] { 0, 0, 0 }, new short[] {
                    0xff, 0xff, 0xff }, "Lucida Grande"));
            minf.add(gmhd);
            break;
        default:
            throw new IllegalStateException("Handler " + type.getHandler() + " not supported");
        }
    }

    private void addDref(NodeBox minf) {
        DataInfoBox dinf = new DataInfoBox();
        minf.add(dinf);
        DataRefBox dref = new DataRefBox();
        dinf.add(dref);
        dref.add(new LeafBox(new Header("alis", 0), new byte[] { 0, 0, 0, 1 }));
    }

    public MuxerTrack getVideoTrack() {
        for (MuxerTrack frameMuxer : tracks) {
            if (frameMuxer.isVideo()) {
                return frameMuxer;
            }
        }
        return null;
    }

    public MuxerTrack getTimecodeTrack() {
        for (MuxerTrack frameMuxer : tracks) {
            if (frameMuxer.isTimecode()) {
                return frameMuxer;
            }
        }
        return null;
    }

    public List<MuxerTrack> getAudioTracks() {
        ArrayList<MuxerTrack> result = new ArrayList<MuxerTrack>();
        for (MuxerTrack frameMuxer : tracks) {
            if (frameMuxer.isAudio()) {
                result.add(frameMuxer);
            }
        }
        return result;
    }

    private void movieHeader(NodeBox movie) {
        int timescale = tracks.get(0).getTimescale();
        long duration = tracks.get(0).getTrackTotalDuration();
        MuxerTrack videoTrack = getVideoTrack();
        if (videoTrack != null) {
            timescale = videoTrack.getTimescale();
            duration = videoTrack.getTrackTotalDuration();
        }

        MovieHeaderBox mvhd = new MovieHeaderBox(timescale, duration, 1.0f, 1.0f, new Date().getTime(),
                new Date().getTime(), new int[] { 0x10000, 0, 0, 0, 0x10000, 0, 0, 0, 0x40000000 }, nextTrackId);
        movie.addFirst(mvhd);
    }

    public static String lookupFourcc(AudioFormat format) {
        if (format.getSampleSizeInBits() == 16 && !format.isBigEndian())
            return "sowt";
        else if (format.getSampleSizeInBits() == 24)
            return "in24";
        else
            throw new IllegalArgumentException("Audio format " + format + " is not supported.");
    }

    public UncompressedTrack addUncompressedAudioTrack(AudioFormat format) {
        return addTrackForUncompressed(SOUND, (int) format.getSampleRate(), 1, (format.getSampleSizeInBits() >> 3)
                * format.getChannels(), MP4Muxer.audioSampleEntry(lookupFourcc(format), 1,
                format.getSampleSizeInBits() >> 3, format.getChannels(), (int) format.getSampleRate(),
                format.isBigEndian() ? Endian.BIG_ENDIAN : Endian.LITTLE_ENDIAN));
    }
}