package org.jcodec.containers.mp4.boxes;

import java.io.DataOutput;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import javax.sound.sampled.AudioFormat;

import org.jcodec.containers.mp4.io.Input;
import org.jcodec.containers.mp4.io.Parser;

/**
 * This class is part of JCodec ( www.jcodec.org ) This software is distributed
 * under FreeBSD License
 * 
 * Describes audio payload sample
 * 
 * @author Stanislav Vitvitskiy
 * 
 */
public class AudioSampleEntry extends SampleEntry {
    private static final MyFactory FACTORY = new MyFactory();
    private short channelCount;
    private short sampleSize;
    private float sampleRate;

    private short revision;
    private int vendor;
    private int compressionId;
    private int pktSize;
    private int samplesPerPkt;
    private int bytesPerPkt;
    private int bytesPerFrame;
    private int bytesPerSample;
    private short version;

    public AudioSampleEntry(Header atom) {
        super(atom);
        factory = FACTORY;
    }

    public AudioSampleEntry(Header header, short drefInd, short channelCount, short sampleSize, int sampleRate,
            short revision, int vendor, int compressionId, int pktSize, int samplesPerPkt, int bytesPerPkt,
            int bytesPerFrame, int bytesPerSample, short version) {
        super(header, drefInd);
        this.channelCount = channelCount;
        this.sampleSize = sampleSize;
        this.sampleRate = sampleRate;
        this.revision = revision;
        this.vendor = vendor;
        this.compressionId = compressionId;
        this.pktSize = pktSize;
        this.samplesPerPkt = samplesPerPkt;
        this.bytesPerPkt = bytesPerPkt;
        this.bytesPerFrame = bytesPerFrame;
        this.bytesPerSample = bytesPerSample;
        this.version = version;
    }

    public void parse(Input input) throws IOException {
        super.parse(input);

        version = (short) Parser.readInt16(input);
        revision = (short) Parser.readInt16(input);
        vendor = (int) Parser.readInt32(input);

        channelCount = (short) Parser.readInt16(input);
        sampleSize = (short) Parser.readInt16(input);

        compressionId = (int) Parser.readInt16(input);
        pktSize = (int) Parser.readInt16(input);

        sampleRate = (float) Parser.readInt32(input) / 65536f;

        if (version == 1) {
            samplesPerPkt = (int) Parser.readInt32(input);
            bytesPerPkt = (int) Parser.readInt32(input);
            bytesPerFrame = (int) Parser.readInt32(input);
            bytesPerSample = (int) Parser.readInt32(input);

            parseExtensions(input);
        }
    }

    protected void doWrite(DataOutput out) throws IOException {
        super.doWrite(out);

        out.writeShort(version);
        out.writeShort(revision);
        out.writeInt(vendor);

        out.writeShort(channelCount);
        if (version == 0)
            out.writeShort(sampleSize);
        else
            out.writeShort(16);

        out.writeShort(compressionId);
        out.writeShort(pktSize);

        out.writeInt((int) Math.round(sampleRate * 65536d));

        if (version == 1) {
            out.writeInt(samplesPerPkt);
            out.writeInt(bytesPerPkt);
            out.writeInt(bytesPerFrame);
            out.writeInt(bytesPerSample);

            writeExtensions(out);
            out.writeInt(0);
        }
    }

    public short getChannelCount() {
        return channelCount;
    }

    public int calcFrameSize() {
        if (version == 0)
            return (sampleSize >> 3) * channelCount;
        else
            return bytesPerFrame;
    }

    public int calcSampleSize() {
        return calcFrameSize() / channelCount;
    }

    public short getSampleSize() {
        return sampleSize;
    }

    public float getSampleRate() {
        return sampleRate;
    }

    public int getBytesPerFrame() {
        return bytesPerFrame;
    }

    public int getBytesPerSample() {
        return bytesPerSample;
    }

    public short getVersion() {
        return version;
    }

    public static class MyFactory extends BoxFactory {
        private Map<String, Class<? extends Box>> mappings = new HashMap<String, Class<? extends Box>>();

        public MyFactory() {
            mappings.put("wave", NodeBox.class);
            mappings.put(FormatBox.fourcc(), FormatBox.class);
            mappings.put(EndianBox.fourcc(), EndianBox.class);
            mappings.put(ChannelBox.fourcc(), ChannelBox.class);
        }

        public Class<? extends Box> toClass(String fourcc) {
            return mappings.get(fourcc);
        }
    }

    public AudioFormat getFormat() {
        return new AudioFormat(sampleRate, calcSampleSize() << 3, channelCount, true, false);
    }
}
