package org.jcodec.containers.mp4.boxes;

import static org.jcodec.containers.mp4.io.Parser.readInt32;
import static org.jcodec.containers.mp4.io.Parser.readInt64;

import java.io.DataOutput;
import java.io.IOException;

import org.jcodec.containers.mp4.io.Input;
import org.jcodec.containers.mp4.io.Parser;

/**
 * This class is part of JCodec ( www.jcodec.org )
 * This software is distributed under FreeBSD License
 * 
 * @author stan
 *
 */
public class TrackHeaderBox extends FullBox {
    private int trackId;
    private long duration;
    private float width;
    private float height;
    private long created;
    private long modified;
    private float volume;
    private short layer;
    private long altGroup;
    private int[] matrix;
    
    public static String fourcc() {
        return "tkhd";
    }

    public TrackHeaderBox(int trackId, long duration, float width, float height, long created,
            long modified, float volume, short layer, long altGroup, int[] matrix) {
        super(new Header(fourcc()));
        this.trackId = trackId;
        this.duration = duration;
        this.width = width;
        this.height = height;
        this.created = created;
        this.modified = modified;
        this.volume = volume;
        this.layer = layer;
        this.altGroup = altGroup;
        this.matrix = matrix;
    }

    public TrackHeaderBox() {
        super(new Header(fourcc()));
    }

    public void parse(Input input) throws IOException {
        super.parse(input);

        if (version == 0) {
            created = readInt32(input); // Creation time
            modified = readInt32(input); // Modification time
        } else {
            created = readInt64(input);
            modified = readInt64(input);
        }
        trackId = (int) readInt32(input);
        readInt32(input);

        if (version == 0) {
            duration = readInt32(input);
        } else {
            duration = readInt64(input);
        }

        Parser.readInt32(input); // Reserved
        Parser.readInt32(input);

        layer = (short)Parser.readInt16(input);
        altGroup = Parser.readInt16(input);

        volume = readVolume(input);

        Parser.readInt16(input);

        readMatrix(input);

        width = Parser.readInt32(input) / 65536f;
        height = Parser.readInt32(input) / 65536f;
    }

    private void readMatrix(Input input) throws IOException {
        matrix = new int[9];
        for (int i = 0; i < 9; i++)
            matrix[i] = (int) Parser.readInt32(input);
    }

    private float readVolume(Input input) throws IOException {
        return (float) (Parser.readInt16(input) / 256.);
    }

    public int getNo() {
        return trackId;
    }

    public long getDuration() {
        return duration;
    }

    public float getWidth() {
        return width;
    }

    public float getHeight() {
        return height;
    }

    public void doWrite(DataOutput out) throws IOException {
        super.doWrite(out);

        out.writeInt((int) created);
        out.writeInt((int) modified);

        out.writeInt(trackId);
        out.writeInt(0);

        out.writeInt((int) duration);

        out.writeInt(0);
        out.writeInt(0);

        out.writeShort((short) layer);
        out.writeShort((short) altGroup);

        writeVolume(out);

        out.writeShort(0);

        writeMatrix(out);

        out.writeInt((int) (width * 65536));
        out.writeInt((int) (height * 65536));
    }

    private void writeMatrix(DataOutput out) throws IOException {
        for (int i = 0; i < 9; i++)
            out.writeInt(matrix[i]);

    }

    private void writeVolume(DataOutput out) throws IOException {
        out.writeShort((short) (volume * 256.));
    }

    public void setDuration(long duration) {
        this.duration = duration;
    }

    public void setNo(int no) {
        this.trackId = no;
    }

    public int[] getMatrix() {
        return matrix;
    }

    public short getLayer() {
        return layer;
    }

    public float getVolume() {
        return volume;
    }

    public void setWidth(float width2) {
        this.width = width2;
    }
    
    public void setHeight(float height) {
        this.height = height;
    }
}