package org.jcodec.containers.mp4.boxes;

import java.io.DataOutput;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.jcodec.codecs.h264.mp4.AvcCBox;
import org.jcodec.containers.mp4.io.Input;
import org.jcodec.containers.mp4.io.Parser;
import org.jcodec.containers.mp4.io.Writer;

/**
 * This class is part of JCodec ( www.jcodec.org )
 * This software is distributed under FreeBSD License
 * 
 * Describes video payload sample
 * 
 * @author Stanislav Vitvitskiy
 * 
 */
public class VideoSampleEntry extends SampleEntry {
    private static final MyFactory FACTORY = new MyFactory();
    private short version;
    private short revision;
    private String vendor;
    private int temporalQual;
    private int spacialQual;
    private long width;
    private long height;
    private float hRes;
    private float vRes;
    private short frameCount;
    private String compressorName;
    private short depth;
    private short clrTbl;

    public VideoSampleEntry(Header atom, short version, short revision, String vendor, int temporalQual,
            int spacialQual, long width, long height, long hRes, long vRes, short frameCount, String compressorName,
            short depth, short drefInd, short clrTbl) {
        super(atom, drefInd);
        factory = FACTORY;
        this.version = version;
        this.revision = revision;
        this.vendor = vendor;
        this.temporalQual = temporalQual;
        this.spacialQual = spacialQual;
        this.width = width;
        this.height = height;
        this.hRes = hRes;
        this.vRes = vRes;
        this.frameCount = frameCount;
        this.compressorName = compressorName;
        this.depth = depth;
        this.clrTbl = clrTbl;
    }

    public VideoSampleEntry(Header atom) {
        super(atom);
        factory = FACTORY;
    }

    public void parse(Input input) throws IOException {
        super.parse(input);

        version = (short) Parser.readInt16(input);
        revision = (short) Parser.readInt16(input);
        vendor = Parser.readString(input, 4);
        temporalQual = (int) Parser.readInt32(input);
        spacialQual = (int) Parser.readInt32(input);

        width = Parser.readInt16(input);
        height = Parser.readInt16(input);

        hRes = (float) Parser.readInt32(input) / 65536f;
        vRes = (float) Parser.readInt32(input) / 65536f;

        Parser.readInt32(input); // Reserved

        frameCount = (short) Parser.readInt16(input);

        compressorName = Parser.readPascalString(input, 31);

        depth = (short) Parser.readInt16(input);

        clrTbl = (short) Parser.readInt16(input);

        parseExtensions(input);
    }

    @Override
    public void doWrite(DataOutput out) throws IOException {

        super.doWrite(out);

        out.writeShort(version);
        out.writeShort(revision);
        out.write(vendor.getBytes(), 0, 4);
        out.writeInt(temporalQual);
        out.writeInt(spacialQual);

        out.writeShort((short) width);
        out.writeShort((short) height);

        out.writeInt((int) (hRes * 65536));
        out.writeInt((int) (vRes * 65536));

        out.writeInt(0); // data size

        out.writeShort(frameCount);

        Writer.writePascalString(out, compressorName, 31);

        out.writeShort(depth);

        out.writeShort(clrTbl);

        writeExtensions(out);
    }

    public long getWidth() {
        return width;
    }

    public long getHeight() {
        return height;
    }

    public float gethRes() {
        return hRes;
    }

    public float getvRes() {
        return vRes;
    }

    public long getFrameCount() {
        return frameCount;
    }

    public String getCompressorName() {
        return compressorName;
    }

    public long getDepth() {
        return depth;
    }

    public static class MyFactory extends BoxFactory {
        private Map<String, Class<? extends Box>> mappings = new HashMap<String, Class<? extends Box>>();

        public MyFactory() {
            mappings.put("pasp", PixelAspectExt.class);
            mappings.put(AvcCBox.fourcc(), AvcCBox.class);
        }

        public Class<? extends Box> toClass(String fourcc) {
            return mappings.get(fourcc);
        }
    }
}